


/**
 * A Robot that completes a task before moving on to the next task, and
 * completes tasks in the order in which they were submitted (using addTask()).
 */
public class NonPreemptiveRobot implements Robot {

	/**
	 * The length of time (in milliseconds) that the robot has actually been
	 * running.
	 */
	private long runningTime;

	/**
	 * The queue of tasks to be run.
	 */
	private EditableQueue<Task> queue;

	public NonPreemptiveRobot() {
		this.runningTime = 0;
		this.queue = new LLEditableQueue<Task>();
	}

	/**
	 * Add a task that the robot needs to execute. The task is added to the end
	 * of the sequence of tasks that the robot is supposed to executed.
	 * 
	 * @param t
	 *            the task to be executed
	 */
	public void addTask(Task t) {
		queue.enqueue(t);
	}

	/**
	 * Removes a task from being run.
	 * 
	 * @param t
	 *            the task to be removed.
	 */
	public void removeTask(Task t) {
		// TODO implement this method
		queue.remove(t);
	}

	/**
	 * Runs the robot for the specified length of time. Note that the duration
	 * specified is only an upper limit for how long the robot can run.
	 * Depending on the instructions that the robot has to execute, it may run
	 * for less than the specified duration. For example, if the next two
	 * instructions to be executed take time 10ms, and 20ms, and the robot is to
	 * run for 15 ms, then there will only be enough time to execute the first
	 * instruction. Note, however, that a robot must make as much progress
	 * executing instructions as it can in the allotted time.
	 * 
	 * See the assignment handout for more clarification on this point.
	 * 
	 * @param duration
	 *            the maximum length of time the robot is allowed to run (in
	 *            milliseconds)
	 */
	public void run(long duration) {

		// TODO implement this method
		
		long remainingDuration = duration;

		// how long the task ran for most recently (initialized to a dummy val)
		long timeDelta = Long.MAX_VALUE;

		/*
		 * Keep running tasks while there is time left and there are more tasks
		 * to be run that can make progress.
		 */
		while (remainingDuration > 0 && timeDelta > 0 && !queue.isEmpty()) {

			// run the task at the front of the queue
			Task curTask = queue.front();
			TaskState oldState = curTask.getState();
			TaskState newState = curTask.run(remainingDuration);

			/*
			 * Update the remaining time that the robot is supposed to run for.
			 * Note that a timeDelta of 0 indicates that the task was unable to
			 * make progress (and hence the robot should stop running).
			 */
			timeDelta = newState.getRunningTime() - oldState.getRunningTime();
			remainingDuration -= timeDelta;

			// if the task is finished, remove it from the queue
			if (newState.isDone()) {
				queue.dequeue();
			}

		}

		// update the total running time of the robot
		runningTime += duration - remainingDuration;
	}

	/**
	 * Retrieves the amount of time the robot has been running for.
	 * 
	 * @return the amount of time the robot has been running for.
	 */
	public long getRunningTime() {
		return runningTime;
	}

}
