


/**
 * A Robot that executes tasks in a round robin manner.
 * 
 * Tasks are stored in a sequence, whose order is initially determined by the
 * order in which they were added to the robot (using the addTask() method).
 * 
 * Robots are supplied with a "quantum" which specifies how long a robot should
 * execute a given task before moving onto the next task. If a task has not been
 * completed by the end of its quantum, then it moves to the end of the sequence
 * of tasks that the robot is executing, and the robot starts executing the next
 * task in the sequence.
 * 
 */
public class RoundRobinRobot implements Robot {

	/**
	 * The length of time (in milliseconds) that the robot has actually been
	 * running.
	 */
	private long runningTime;

	/**
	 * The queue of tasks to be run.
	 */
	private EditableQueue<Task> queue;

	/**
	 * The time that the robot should spend on a given task before moving to the
	 * next one.
	 */
	private long quantum;

	/**
	 * Constructs a new RoundRobinRobot.
	 * 
	 * @param quantum
	 *            the time that a robot should spend on a given task before
	 *            switching to the next one.
	 */
	public RoundRobinRobot(long quantum) {
		this.runningTime = 0;
		this.queue = new LLEditableQueue<Task>();
		this.quantum = quantum;
	}

	/**
	 * Add a task that the robot needs to execute. The task is added to the end
	 * of the sequence of tasks that the robot is supposed to executed.
	 * 
	 * @param t
	 *            the task to be executed
	 */
	public void addTask(Task t) {
		queue.enqueue(t);
	}

	/**
	 * Removes a task from being run.
	 * 
	 * @param t
	 *            the task to be removed.
	 */
	public void removeTask(Task t) {
		// TODO implement this method
		queue.remove(t);
	}

	/**
	 * Runs the robot for the specified length of time. Note that the duration
	 * specified is only an upper limit for how long the robot can run.
	 * Depending on the instructions that the robot has to execute, it may run
	 * for less than the specified duration. For example, if the next two
	 * instructions to be executed take time 10ms, and 20ms, and the robot is to
	 * run for 15 ms, then there will only be enough time to execute the first
	 * instruction. Note, however, that a robot must make as much progress
	 * executing instructions as it can in the allotted time.
	 * 
	 * See the assignment handout for more clarification on this point.
	 * 
	 * @param duration
	 *            the maximum length of time the robot is allowed to run.
	 */
	public void run(long duration) {
		
		// TODO implement this method
		
		long remainingDuration = duration;

		/*
		 * Keep running tasks while there is time left and there is tasks left
		 * to be run
		 */
		while (remainingDuration >= quantum && !queue.isEmpty()) {

			// run the task for the specified quantum
			Task curTask = queue.dequeue();
			TaskState oldState = curTask.getState();
			TaskState newState = curTask.run(quantum);

			// subtract from the remaining time the time that the task actually
			// ran for
			remainingDuration -= newState.getRunningTime()
					- oldState.getRunningTime();

			// if the task isn't done, then it needs to be enqueued again
			if (!newState.isDone()) {
				queue.enqueue(curTask);
			}
		}

		// update the total running time of the robot
		runningTime += duration - remainingDuration;
	}

	/**
	 * Retrieves the total amount of time the robot has been running for.
	 * 
	 * @return the total amount of time the robot has been running for.
	 */
	public long getRunningTime() {
		return runningTime;
	}

}
