


/**
 * A task is essentially a "program" that a Robot can execute.
 * 
 */
public class Task {

	/**
	 * The function call stack. The top of the stack contains the function
	 * currently being executed.
	 */
	private PureStack<Function> callStack;

	/**
	 * Stores the index of the next instruction to be executed by the function
	 * at the same level in the callStack.
	 */
	private PureStack<Integer> indexStack;

	/**
	 * The current state of the task.
	 */
	private TaskState curState;

	/**
	 * Constructs a new Task.
	 * 
	 * @param main
	 *            the function where execution of the task should begin.
	 */
	public Task(Function main) {
		callStack = new LLPureStack<Function>();
		indexStack = new LLPureStack<Integer>();

		/*
		 * Initialize the call stack and index stack to indicate that the
		 * instruction at index 0 in the main function is where execution should
		 * begin.
		 */
		callStack.push(main);
		indexStack.push(0);

		curState = new TaskState(main, 0, 0, false);
	}

	/**
	 * Runs the task for the specified duration. Note that the specified
	 * duration is an upper limit as to how long the task should run. Depending
	 * on the execution time of individual instructions, the task may run for
	 * less than the specified duration. For example, if the next instruction
	 * takes 10ms to execute, but the task is to be run for 5ms, then there's
	 * not enough time to execute the instruction and the task won't run at all.
	 * If there is enough time to execute the instruction, however, then the
	 * instruction will be executed.
	 * 
	 * @param duration
	 *            the maximum length of time (in milliseconds) that the task to
	 *            run for
	 * @return the new state of the task after having ran the task for the
	 *         specified duration
	 */
	public TaskState run(long duration) {

		// TODO implement this method
		
		// the amount of time left for executing this task
		long remainingDuration = duration;

		// is there enough time left to execute another instruction?
		boolean isEnoughTimeLeft = true;

		/*
		 * Keep executing instructions while there's time available to do so and
		 * a return instruction hasn't been encountered in the main function.
		 */
		while (isEnoughTimeLeft && !callStack.isEmpty()) {
			Function curFunction = callStack.peek();
			Instruction curInstruction = curFunction.getInstruction(indexStack
					.peek());

			if (remainingDuration >= curInstruction.getExecuteTime()) {
				remainingDuration -= curInstruction.getExecuteTime();
				handleInstruction(curInstruction);
			} else {
				isEnoughTimeLeft = false;
			}

		}

		// update the current state of the task
		long newTime = curState.getRunningTime()
				+ (duration - remainingDuration);
		curState = new TaskState(callStack.peek(),
				(indexStack.peek() != null ? indexStack.peek() : -1), newTime,
				callStack.peek() == null);

		return curState;
	}

	/**
	 * Handle the current instruction being executed. The semantics of how this
	 * is done depends on the instruction type.
	 * 
	 * @param curInstruction
	 *            the instruction currently being executed
	 */
	private void handleInstruction(Instruction curInstruction) {
		switch (curInstruction.getType()) {
		case Instruction.CALL_TYPE: {

			/*
			 * A function is being called. So in this case we first increment
			 * the index of the next instruction to be executed in the current
			 * function. We then push the function being called onto the
			 * callStack, and then push 0 onto the indexStack to indicate that
			 * the instruction at index 0 is where the execution should begin in
			 * the called function.
			 */
			int index = indexStack.pop();
			indexStack.push(++index);

			Object[] operands = curInstruction.getOperands();
			callStack.push((Function) operands[0]);
			indexStack.push(0);

			break;
		}
		case Instruction.RETURN_TYPE:

			/*
			 * A return to the caller is made from the current function. All we
			 * need to do is pop each stack.
			 */
			callStack.pop();
			indexStack.pop();
			break;
		case Instruction.GOTO_TYPE: {

			/*
			 * The goto instruction specifies the index of the instruction in
			 * the current function that should be next executed. Update the
			 * indexStack to reflect this.
			 */
			Object[] operands = curInstruction.getOperands();
			indexStack.pop();
			indexStack.push((Integer) operands[0]);
			break; 
		}
		default:
			/*
			 * In this case the instruction type is OTHER_TYPE. Here all we do
			 * increment the index of the instruction to be executed.
			 */
			int index = indexStack.pop();
			indexStack.push(++index);

			break;
		}
	}

	/**
	 * Retrieves the state that the task is currently in. It contains
	 * information about the function currently executing, the index of the
	 * instruction to be next executed, and the actual running time of the task.
	 * 
	 * @return the current state of the task.
	 */
	public TaskState getState() {
		return curState;
	}

}
