import junit.framework.TestCase;

public class TaskTester extends TestCase {

	/**
	 * Simple test involves running a function that contains only RETURN.
	 * Running for 0 time.
	 * 
	 * 
	 */
	public void testRun1() {
		Instruction m0 = new ReturnInstruction();

		Function main = new Function();
		main.addInstruction(m0);

		Task t = new Task(main);

		TaskState ts = t.run(0);

		assertFalse(ts.isDone());
		assertEquals(main, ts.getCurrentFunction());
		assertEquals(0, ts.getInstructionIndex());
		assertEquals(0, ts.getRunningTime());

	}

	/**
	 * Simple test involves running a function that contains only RETURN. Run
	 * for less time than needed.
	 * 
	 */
	public void testRun2() {
		Instruction m0 = new ReturnInstruction();

		Function main = new Function();
		main.addInstruction(m0);

		Task t = new Task(main);

		TaskState ts = t.run(5);
		TaskState ts2 = t.getState();
		assertEquals(ts, ts2);

		assertFalse(ts.isDone());
		assertEquals(main, ts.getCurrentFunction());
		assertEquals(0, ts.getInstructionIndex());
		assertEquals(0, ts.getRunningTime());
	}

	/**
	 * Simple test involves running a function that contains only RETURN. Run
	 * for exact time needed.
	 * 
	 */
	public void testRun3() {
		Instruction m0 = new ReturnInstruction();

		Function main = new Function();
		main.addInstruction(m0);

		Task t = new Task(main);

		TaskState ts = t.run(ReturnInstruction.EXECUTE_TIME);
		assertTrue(ts.isDone());
		assertEquals(ReturnInstruction.EXECUTE_TIME, ts.getRunningTime());

		// running after task is done should have no effect
		TaskState ts2 = t.run(20);
		assertEquals(ts, ts2);
	}

	/**
	 * Simple test involves running a function that contains only RETURN. Run
	 * for more time than needed.
	 * 
	 */
	public void testRun4() {
		Instruction m0 = new ReturnInstruction();

		Function main = new Function();
		main.addInstruction(m0);

		Task t = new Task(main);

		TaskState ts = t.run(27);
		assertTrue(ts.isDone());
		assertEquals(ReturnInstruction.EXECUTE_TIME, ts.getRunningTime());

		// running after task is done should have no effect
		TaskState ts2 = t.run(ReturnInstruction.EXECUTE_TIME);
		assertEquals(ts, ts2);
	}

	/**
	 * More complicated test that involves calling a function.
	 * 
	 * function main:
	 * 
	 * 0 OTHER_TYPE 15ms 
	 * 1 CALL_TYPE B 10ms 
	 * 2 CALL_TYPE C 10ms 
	 * 3 RETURN 10ms
	 * 
	 * function B: 
	 * 0 CALL_TYPE C 10ms 
	 * 1 RETURN_TYPE 10ms
	 * 
	 * function C: 
	 * 0 RETURN_TYPE 10ms
	 * 
	 */
	public void testRun5() {
		Instruction c0 = new ReturnInstruction();
		Function C = new Function();
		C.addInstruction(c0);

		Instruction b0 = new CallInstruction(C);
		Instruction b1 = new ReturnInstruction();
		Function B = new Function();
		B.addInstruction(b0);
		B.addInstruction(b1);

		Instruction m0 = new OtherInstruction(15);
		Instruction m1 = new CallInstruction(B);
		Instruction m2 = new CallInstruction(C);
		Instruction m3 = new ReturnInstruction();

		Function main = new Function();
		main.addInstruction(m0);
		main.addInstruction(m1);
		main.addInstruction(m2);
		main.addInstruction(m3);

		Task t = new Task(main);

		TaskState ts = t.run(10);
		check(ts, false, 0, main, 0);

		ts = t.run(15);
		check(ts, false, 15, main, 1);
		
		ts = t.run(45); 
		check(ts, false, 55, main, 2);
		
		ts = t.run(10); 
		check(ts, false, 65, C, 0);
		
		ts = t.run(19); 
		check(ts, false, 75, main, 3); 
		
		ts = t.run(10); 
		check(ts, true, 85, main, -1); 
	}

	/**
	 * Performs assertions for checking the TaskState was set properly. 
	 * 
	 * @param ts task state
	 * @param done whether task should be done 
	 * @param runningTime the current running time of the task 
	 * @param curFunction the current function executing 
	 * @param index the instruction to be next executed
	 */
	private void check(TaskState ts, boolean done, long runningTime,
			Function curFunction, int index) {
		assertEquals(done, ts.isDone());
		assertEquals(runningTime, ts.getRunningTime());
		if (!done) {
			assertEquals(curFunction, ts.getCurrentFunction());
			assertEquals(index, ts.getInstructionIndex());
		}
	}
	
	
	/**
	 * Test that involves deep call chain
	 * 
	 * Function main:
	 * 0  CALL B   10ms
	 * 1  RETURN   10ms 
	 *
	 * Function B
	 * 0 CALL C    10ms
	 * 1 RETURN    10ms 
	 * 
	 * Function C 
	 * 0 CALL D    10ms
	 * 1 RETURN    10ms
	 * 
	 * Function D
	 * 0 RETURN    10ms 
	 */
	public void testRun6() {
		Instruction d0 = new ReturnInstruction(); 
		Function D = new Function(); 
		D.addInstruction(d0);
		
		Instruction c0 = new CallInstruction(D);
		Instruction c1 = new ReturnInstruction(); 
		Function C = new Function(); 
		C.addInstruction(c0);
		C.addInstruction(c1); 
		
		Instruction b0 = new CallInstruction(C);
		Instruction b1 = new ReturnInstruction(); 
		Function B = new Function(); 
		B.addInstruction(b0);
		B.addInstruction(b1);
		
		Instruction m0 = new CallInstruction(B); 
		Instruction m1 = new ReturnInstruction(); 
		Function main = new Function(); 
		main.addInstruction(m0); 
		main.addInstruction(m1); 
		
		Task t = new Task(main); 
		
		TaskState ts = t.run(29); 
		check(ts, false, 20, C, 0); 
		
		ts = t.run(30); 
		check(ts, false, 50, B, 1); 
		
		ts = t.run(500); 
		check(ts, true, 70, main, -1); 


	}
	
	/**
	 * Test that involve GOTO skipping some instructions 
	 * 
	 * Function main:
	 * 0  GOTO_TYPE 2  20 ms
	 * 1  RETURN       10 ms
	 * 2  GOTO_TYPE 4  20 ms
	 * 3  RETURN       10 ms
	 * 4  GOTO_TYPE 1  20 ms 
	 */
	public void testRun7() {
		Instruction m0 = new GotoInstruction(2); 
		Instruction m1 = new ReturnInstruction(); 
		Instruction m2 = new GotoInstruction(4); 
		Instruction m3 = new ReturnInstruction(); 
		Instruction m4 = new GotoInstruction(1); 
		
		Function main = new Function(); 
		main.addInstruction(m0); 
		main.addInstruction(m1); 
		main.addInstruction(m2); 
		main.addInstruction(m3); 
		main.addInstruction(m4);
		
		Task t = new Task(main); 
		TaskState ts = t.run(25);
		
		check(ts, false, 20, main, 2); 
		
		ts = t.run(19); 
		check(ts, false, 20, main, 2); 
		
		ts = t.run(40); 
		check(ts, false, 60, main, 1); 
		
		ts = t.run(10); 
		check(ts, true, 70, main, -1); 
		
		
	}
	
	/**
	 * Test case with infinite loop 
	 * 
	 * Function main
	 * 0   GOTO_TYPE 2  20 ms
	 * 1   OTHER_TYPE   17 ms
	 * 2   GOTO_TYPE 1  20 ms 
	 */
	public void testRun8() {
		Instruction m0 = new GotoInstruction(2); 
		Instruction m1 = new OtherInstruction(17); 
		Instruction m2 = new GotoInstruction(1); 
		
		Function main = new Function(); 
		main.addInstruction(m0); 
		main.addInstruction(m1); 
		main.addInstruction(m2); 
		
		Task t = new Task(main); 
		TaskState ts = t.run(20);
		
		check(ts, false, 20, main, 2); 
		
		ts = t.run(20); 
		check(ts, false, 40, main, 1); 
		
		ts = t.run(37); 
		check(ts, false, 77, main, 1); 
		
		ts = t.run(36); 
		check(ts, false, 94, main, 2); 
		
		ts = t.run(20); 
		check(ts, false, 114, main, 1); 
		
		ts = t.run(37*10000); 
		check(ts, false, 114 + 37 * 10000, main, 1); 
		
	}
	

}
